from typing import Any, Optional, Set, Dict

import pydantic_core
from pydantic import BaseModel, PrivateAttr, ValidationError, TypeAdapter
from pydantic._internal import _fields

from lazy_model.nao import NAO

ROOT_KEY = "__root__"
_object_setattr = object.__setattr__


class LazyModel(BaseModel):
    _store: Dict[str, Any] = PrivateAttr(default_factory=dict)
    _lazily_parsed: bool = PrivateAttr(default=False)

    @classmethod
    def lazy_parse(
        cls,
        data: Dict[str, Any],
        fields: Optional[Set[str]] = None,
    ):
        fields_values = {}
        field_alias_map = {}
        if fields is None:
            fields = set()
        for name, field in cls.model_fields.items():
            fields_values[name] = NAO
            alias = field.alias or name
            if alias in fields:
                field_alias_map[alias] = name
        field_set = set(fields_values.keys())
        m = cls.model_construct(field_set, **fields_values)
        m._store = data

        for alias in fields:
            m._set_attr(field_alias_map[alias], data[alias])
        return m

    def _parse_value(self, name, value):
        field_type = self.model_fields.get(name).annotation
        try:
            value = TypeAdapter(field_type).validate_python(value)
        except ValidationError as e:
            if value is None and self.model_fields.get(name).required is False:
                value = None
            else:
                raise e
        return value

    def parse_store(self):
        for name in self.model_fields:
            self.__getattribute__(name)

    def _set_attr(self, name: str, value: Any) -> None:
        """
        Stolen from Pydantic.
        :param name:
        :param value:
        :return:

        TODO rework
        """
        if name in self.__class_vars__:
            raise AttributeError(
                f"{name!r} is a ClassVar of `{self.__class__.__name__}` and cannot be set on an instance. "  # noqa: E501
                f"If you want to set a value on the class, use `{self.__class__.__name__}.{name} = value`."  # noqa: E501
            )
        elif not _fields.is_valid_field_name(name):
            if (
                self.__pydantic_private__ is None
                or name not in self.__private_attributes__
            ):
                _object_setattr(self, name, value)
            else:
                attribute = self.__private_attributes__[name]
                if hasattr(attribute, "__set__"):
                    attribute.__set__(self, value)  # type: ignore
                else:
                    self.__pydantic_private__[name] = value
            return
        elif self.model_config.get("frozen", None):
            error: pydantic_core.InitErrorDetails = {
                "type": "frozen_instance",
                "loc": (name,),
                "input": value,
            }
            raise pydantic_core.ValidationError.from_exception_data(
                self.__class__.__name__, [error]
            )

        attr = getattr(self.__class__, name, None)
        if isinstance(attr, property):
            attr.__set__(self, value)
        self.__pydantic_validator__.validate_assignment(self, name, value)

    def __getattribute__(self, item):
        res = super(LazyModel, self).__getattribute__(item)
        if res is NAO:
            field_info = self.model_fields.get(item)
            alias = field_info.alias or item
            value = self._store.get(alias, NAO)
            if value is NAO:
                value = field_info.get_default()
            else:
                value = self._parse_value(item, value)
            self._set_attr(item, value)
            res = super(LazyModel, self).__getattribute__(item)
        return res



from typing import Any, Optional, Set, Dict

from pydantic import BaseModel, PrivateAttr, parse_obj_as, ValidationError

from pydantic.error_wrappers import ErrorWrapper

from lazy_model.nao import NAO

ROOT_KEY = "__root__"
object_setattr = object.__setattr__


class LazyModel(BaseModel):
    _store: Dict[str, Any] = PrivateAttr(default_factory=dict)
    _lazily_parsed: bool = PrivateAttr(default=False)

    @classmethod
    def lazy_parse(
        cls,
        data: Dict[str, Any],
        fields: Optional[Set[str]] = None,
        **new_kwargs,
    ):
        m = cls.__new__(cls, **new_kwargs)  # type: ignore
        if fields is None:
            fields = set()
        fields_values: Dict[str, Any] = {}
        field_alias_map: Dict[str, str] = {}
        for name, field in cls.__fields__.items():
            fields_values[name] = NAO
            if field.alias in fields:
                field_alias_map[field.alias] = name
        object_setattr(m, "__dict__", fields_values)
        _fields_set = set(fields_values.keys())
        object_setattr(m, "__fields_set__", _fields_set)
        m._init_private_attributes()

        m._store = data
        m._lazily_parsed = True

        for alias in fields:
            m._set_attr(field_alias_map[alias], data[alias])
        return m

    def _set_attr(self, name, value):
        """
        Stolen from Pydantic.

        # TODO rework
        """
        field_type = self.__fields__.get(name).annotation
        try:
            value = parse_obj_as(field_type, value)
        except ValidationError as e:
            if value is None and self.__fields__.get(name).required is False:
                value = None
            else:
                raise e

        new_values = {**self.__dict__, name: value}

        for validator in self.__pre_root_validators__:
            try:
                new_values = validator(self.__class__, new_values)
            except (ValueError, TypeError, AssertionError) as exc:
                raise ValidationError(
                    [ErrorWrapper(exc, loc=ROOT_KEY)], self.__class__
                )

        known_field = self.__fields__.get(name, None)
        if known_field:
            if not known_field.field_info.allow_mutation:
                raise TypeError(
                    f'"{known_field.name}" has allow_mutation set '
                    f"to False and cannot be assigned"
                )
            dict_without_original_value = {
                k: v for k, v in self.__dict__.items() if k != name
            }
            value, error_ = known_field.validate(
                value,
                dict_without_original_value,
                loc=name,
                cls=self.__class__,
            )
            if error_:
                raise ValidationError([error_], self.__class__)
            else:
                new_values[name] = value

        errors = []
        for skip_on_failure, validator in self.__post_root_validators__:
            if skip_on_failure and errors:
                continue
            try:
                new_values = validator(self.__class__, new_values)
            except (ValueError, TypeError, AssertionError) as exc:
                errors.append(ErrorWrapper(exc, loc=ROOT_KEY))
        if errors:
            raise ValidationError(errors, self.__class__)

        object_setattr(self, "__dict__", new_values)

        self.__fields_set__.add(name)

    def parse_store(self):
        for name in self.__fields__:
            self.__getattribute__(name)

    def __getattribute__(self, item):
        res = super(LazyModel, self).__getattribute__(item)
        if res is NAO:
            field_info = self.__fields__.get(item)
            value = self._store.get(field_info.alias, NAO)
            if value is NAO:
                value = field_info.get_default()
            self._set_attr(item, value)
            res = super(LazyModel, self).__getattribute__(item)
        return res



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lazy_model', 'lazy_model.parser']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.0']

setup_kwargs = {
    'name': 'lazy-model',
    'version': '0.1.0b0',
    'description': '',
    'long_description': '# Lazy parsing for Pydantic models\n\nThis library provides a lazy interface for parsing objects from dictionaries. During the parsing, it saves the raw data inside the object and parses each field on demand.\n\n## Install\n\npoetry\n```shell\npoetry add lazy-model\n```\n\npip\n```shell\npip install lazy-model\n```\n\n## Usage\n\n```python\nfrom lazy_model import LazyModel\nfrom pydantic import validator\n\n\nclass Sample(LazyModel):\n    i: int\n    s: str\n\n    @validator("s")\n    def s_upper(cls, v):\n        return v.upper()\n\n\nobj = Sample.lazy_parse({"i": "10", "s": "test"})\n\n# at this point the data is stored in a raw format inside the object\n\nprint(obj.__dict__)\n\n# >>> {\'i\': NAO, \'s\': NAO}\n\n# NAO - Not An Object. It shows that the field was not parsed yet.\n\nprint(obj.s)\n\n# >>> TEST\n\n# Custom validator works during lazy parsing\n\nprint(obj.__dict__)\n\n# >>> {\'i\': NAO, \'s\': \'TEST\'}\n\n# The `s` field  was already parsed by this step\n\nprint(obj.i, type(obj.i))\n\n# >>> 10 <class \'int\'>\n\n# It converted `10` from string to int based on the annotations\n\nprint(obj.__dict__)\n\n# >>> {\'i\': 10, \'s\': \'TEST\'}\n\n# Everything was parsed\n```',
    'author': 'Roman Right',
    'author_email': 'roman-right@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
